<!doctype html>
<html lang="en">
  <head>
    <meta charset="UTF-8" />
    <link rel="icon" type="image" href="/zndraw.png" />
    <meta name="viewport" content="width=device-width, initial-scale=1.0" />
    <title>ZnDraw</title>
    <script type="module" crossorigin src="/assets/index-DNYIA-rP.js"></script>
    <link rel="stylesheet" crossorigin href="/assets/index-C7GIUsX9.css">
  </head>
  <body>
    <div id="root"></div>
  </body>
</html>
